from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict, Field, RootModel


class ItemPayload(RootModel[dict[str, Any]]):
    pass


class SelectionScopePayload(BaseModel):
    q: str | None = None
    filters: dict[str, str] = Field(default_factory=dict)


class IdsPayload(BaseModel):
    ids: list[Any] = Field(default_factory=list)
    select_all: bool = False
    selection_scope: SelectionScopePayload | None = None


class BulkActionPayload(BaseModel):
    ids: list[Any] = Field(default_factory=list)
    select_all: bool = False
    selection_scope: SelectionScopePayload | None = None
    model_config = ConfigDict(extra="allow")

    @property
    def action_payload(self) -> dict[str, Any]:
        return {
            key: value
            for key, value in self.model_dump().items()
            if key not in {"ids", "select_all", "selection_scope"}
        }


class PaginationPayload(BaseModel):
    limit: int
    offset: int
    total: int


class ChoiceItemPayload(BaseModel):
    id: str | int
    label: str


class ChoicesResponse(BaseModel):
    items: list[ChoiceItemPayload]


class ModelsResponse(BaseModel):
    locale: str
    items: list[dict[str, Any]]
    blocks: list[dict[str, Any]]


class ModelResponse(RootModel[dict[str, Any]]):
    pass


class ListResponse(BaseModel):
    meta: dict[str, Any]
    pagination: PaginationPayload
    items: list[dict[str, Any]]


class DetailResponse(BaseModel):
    meta: dict[str, Any]
    item: dict[str, Any]


class ItemOnlyResponse(BaseModel):
    item: dict[str, Any]


class DeleteCountsPayload(BaseModel):
    roots: int
    delete: int
    protect: int
    set_null: int
    total: int


class DeletePreviewNodePayload(BaseModel):
    model_slug: str | None
    model_title: str
    relation_name: str | None
    id: str | int
    label: str
    effect: str
    children: list[DeletePreviewNodePayload]


DeletePreviewNodePayload.model_rebuild()


class DeletePreviewResponse(BaseModel):
    can_delete: bool
    summary: DeleteCountsPayload
    roots: list[DeletePreviewNodePayload]


class DeleteResultResponse(BaseModel):
    deleted: int


class ProcessedResponse(BaseModel):
    processed: int


class FlexibleProcessedResponse(BaseModel):
    processed: int
    model_config = ConfigDict(extra="allow")


class ObjectActionResponse(BaseModel):
    item: dict[str, Any]
    result: dict[str, Any]



from __future__ import annotations

from datetime import date

import pytest
from sqlalchemy import select

from ._router_helpers import (
    DemoInviteORM,
    DemoRoleORM,
    DemoUserORM,
    build_test_app,
    get_test_client,
    save_entities,
)

pytestmark = pytest.mark.asyncio


async def test_router_supports_virtual_fields_and_custom_setter() -> None:
    app, session_factory = await build_test_app()

    async with get_test_client(app) as client:
        create_response = await client.post(
            "/xladmin/models/users/items/",
            json={"username": "tester", "joined_on": "2026-03-26", "new_password": "secret"},
        )

        assert create_response.status_code == 201
        assert create_response.json()["item"]["new_password"] == ""

    async with session_factory() as session:
        item = (await session.execute(select(DemoUserORM))).scalar_one()
        assert item.password == "hashed::secret"


async def test_router_uses_create_item_factory_for_hidden_required_fields() -> None:
    app, session_factory = await build_test_app()

    async with get_test_client(app) as client:
        create_response = await client.post(
            "/xladmin/models/invites/items/",
            json={"email": "invite@example.com"},
        )

        assert create_response.status_code == 201
        assert create_response.json()["item"]["email"] == "invite@example.com"

    async with session_factory() as session:
        item = (await session.execute(select(DemoInviteORM))).scalar_one()
        assert item.secret_key == "generated-secret"
        assert item.created_by == "admin"


async def test_router_returns_400_when_hidden_required_create_fields_are_missing() -> None:
    app, _session_factory = await build_test_app()

    async with get_test_client(app) as client:
        create_response = await client.post(
            "/xladmin/models/strict-profiles/items/",
            json={"name": "alpha"},
        )

        assert create_response.status_code == 400
        assert create_response.json()["detail"] == "Missing required fields for create: access_code."


async def test_router_parses_string_boolean_payloads_correctly() -> None:
    app, session_factory = await build_test_app()

    async with get_test_client(app) as client:
        create_response = await client.post(
            "/xladmin/models/users/items/",
            json={"username": "tester", "is_active": "false"},
        )

        assert create_response.status_code == 201
        assert create_response.json()["item"]["is_active"] is False

    async with session_factory() as session:
        item = (await session.execute(select(DemoUserORM))).scalar_one()
        assert item.is_active is False


async def test_router_rejects_unknown_relation_ids() -> None:
    app, _session_factory = await build_test_app()

    async with get_test_client(app) as client:
        response = await client.post("/xladmin/models/audits/items/", json={"role_id": 999})

        assert response.status_code == 400
        assert response.json()["detail"] == "Invalid value for field 'role_id'."


async def test_router_supports_custom_query_for_list() -> None:
    app, session_factory = await build_test_app()
    await save_entities(session_factory, DemoRoleORM(name="admin"), DemoRoleORM(name="hidden"))

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/roles/items/")

        assert response.status_code == 200
        assert [item["name"] for item in response.json()["items"]] == ["admin"]


async def test_router_applies_query_for_list_scope_to_detail_endpoints() -> None:
    app, session_factory = await build_test_app()
    await save_entities(session_factory, DemoRoleORM(name="admin"), DemoRoleORM(name="hidden"))

    async with get_test_client(app) as client:
        detail_response = await client.get("/xladmin/models/roles/items/2/")
        delete_response = await client.delete("/xladmin/models/roles/items/2/")

        assert detail_response.status_code == 404
        assert delete_response.status_code == 404


async def test_router_uses_list_display_and_custom_search() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoUserORM(username="alpha", password="hashed::1"),
        DemoUserORM(username="beta", password="hashed::2"),
    )

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/users/items/", params={"q": "al"})

        assert response.status_code == 200
        payload = response.json()
        display_name_meta = next(field for field in payload["meta"]["fields"] if field["name"] == "display_name")
        assert payload["meta"]["list_fields"] == ["id", "display_name", "joined_on", "is_active"]
        assert payload["meta"]["page_size"] == 120
        assert display_name_meta["width_px"] == 320
        assert payload["items"][0]["display_name"].endswith("alpha")
        assert payload["pagination"]["total"] == 1


async def test_router_supports_bulk_actions() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoUserORM(username="alpha", password="hashed::1", is_active=False),
        DemoUserORM(username="beta", password="hashed::2", is_active=False),
    )

    async with get_test_client(app) as client:
        response = await client.post("/xladmin/models/users/bulk-actions/activate/", json={"ids": [1, 2]})

        assert response.status_code == 200
        assert response.json()["processed"] == 2
        assert response.json()["activated"] == 2

    async with session_factory() as session:
        items = list((await session.execute(select(DemoUserORM).order_by(DemoUserORM.id))).scalars())
        assert [item.is_active for item in items] == [True, True]


async def test_router_supports_bulk_actions_for_all_filtered_items() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoUserORM(username="alpha", password="hashed::1", is_active=False),
        DemoUserORM(username="beta", password="hashed::2", is_active=False),
        DemoUserORM(username="gamma", password="hashed::3", is_active=True),
    )

    async with get_test_client(app) as client:
        response = await client.post(
            "/xladmin/models/users/bulk-actions/activate/",
            json={
                "ids": [],
                "select_all": True,
                "selection_scope": {
                    "filters": {"status": "false"},
                },
            },
        )

        assert response.status_code == 200
        assert response.json()["processed"] == 2
        assert response.json()["activated"] == 2

    async with session_factory() as session:
        items = list((await session.execute(select(DemoUserORM).order_by(DemoUserORM.id))).scalars())
        assert [item.is_active for item in items] == [True, True, True]


async def test_router_supports_sorting_by_multiple_fields() -> None:
    app, session_factory = await build_test_app()
    await save_entities(
        session_factory,
        DemoUserORM(username="beta", password="hashed::1", joined_on=date(2026, 3, 27), is_active=False),
        DemoUserORM(username="alpha", password="hashed::2", joined_on=date(2026, 3, 25), is_active=True),
        DemoUserORM(username="gamma", password="hashed::3", joined_on=date(2026, 3, 26), is_active=False),
    )

    async with get_test_client(app) as client:
        by_display_response = await client.get("/xladmin/models/users/items/", params={"sort": "display_name"})
        by_multiple_fields_response = await client.get("/xladmin/models/users/items/", params={"sort": "is_active,-id"})
        by_date_response = await client.get("/xladmin/models/users/items/", params={"sort": "joined_on"})

        assert by_display_response.status_code == 200
        assert [item["id"] for item in by_display_response.json()["items"]] == [2, 1, 3]

        assert by_multiple_fields_response.status_code == 200
        assert [item["id"] for item in by_multiple_fields_response.json()["items"]] == [3, 1, 2]

        assert by_date_response.status_code == 200
        assert [item["joined_on"] for item in by_date_response.json()["items"]] == [
            "2026-03-25",
            "2026-03-26",
            "2026-03-27",
        ]


async def test_router_supports_object_actions() -> None:
    app, session_factory = await build_test_app()
    await save_entities(session_factory, DemoUserORM(username="alpha", password="hashed::1", is_active=True))

    async with get_test_client(app) as client:
        meta_response = await client.get("/xladmin/models/users/")
        action_response = await client.post("/xladmin/models/users/items/1/actions/deactivate/", json={})

        assert meta_response.status_code == 200
        assert meta_response.json()["object_actions"] == [{"slug": "deactivate", "label": "Деактивировать"}]

        assert action_response.status_code == 200
        assert action_response.json()["result"]["deactivated"] == 1
        assert action_response.json()["item"]["is_active"] is False


async def test_router_returns_english_locale_and_builtin_delete_label() -> None:
    app, _session_factory = await build_test_app(locale="en")

    async with get_test_client(app) as client:
        response = await client.get("/xladmin/models/")

        assert response.status_code == 200
        payload = response.json()
        assert payload["locale"] == "en"
        assert payload["items"][0]["locale"] == "en"
        assert payload["items"][0]["bulk_actions"][0] == {"slug": "delete", "label": "Delete"}


async def test_router_returns_locale_in_model_list_detail_and_item_meta() -> None:
    app, session_factory = await build_test_app(locale="en")
    await save_entities(session_factory, DemoUserORM(username="alpha", password="hashed::1"))

    async with get_test_client(app) as client:
        model_response = await client.get("/xladmin/models/users/")
        list_response = await client.get("/xladmin/models/users/items/")
        item_response = await client.get("/xladmin/models/users/items/1/")

        assert model_response.status_code == 200
        assert model_response.json()["locale"] == "en"

        assert list_response.status_code == 200
        assert list_response.json()["meta"]["locale"] == "en"

        assert item_response.status_code == 200
        assert item_response.json()["meta"]["locale"] == "en"



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Literal

from sqlalchemy import inspect as sa_inspect
from sqlalchemy import or_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import RelationshipDirection

from xladmin.config import ModelConfig
from xladmin.introspection import get_display_value, get_pk_value
from xladmin.registry import Registry
from xladmin.serializer import serialize_scalar

DeleteEffect = Literal["delete", "protect", "set-null"]
NodeKey = tuple[type[Any], Any]
_DEPENDENT_RELATIONS_CACHE: dict[type[Any], list[_DependentRelation]] = {}


async def build_delete_plan(
        session: AsyncSession,
        registry: Registry,
        model_config: ModelConfig,
        items: list[Any],
) -> tuple[dict[str, Any], list[Any], list[tuple[Any, list[str]]]]:
    ordered_items: list[Any] = []
    ordered_keys: set[NodeKey] = set()
    set_null_items: list[tuple[Any, list[str]]] = []
    set_null_index: dict[NodeKey, int] = {}
    edges_by_parent = await _build_dependency_graph(
        session=session,
        registry=registry,
        root_model_config=model_config,
        root_items=items,
        set_null_items=set_null_items,
        set_null_index=set_null_index,
    )
    roots = [
        _build_preview_node(
            registry=registry,
            model_config=model_config,
            item=item,
            relation_name=None,
            effect="delete",
            path_seen=set(),
            edges_by_parent=edges_by_parent,
            ordered_items=ordered_items,
            ordered_keys=ordered_keys,
        )
        for item in items
    ]
    counts = _count_effects(roots)
    preview = {
        "can_delete": counts["protect"] == 0,
        "summary": {
            "roots": len(roots),
            "delete": counts["delete"],
            "protect": counts["protect"],
            "set_null": counts["set-null"],
            "total": len(roots) + counts["delete"] + counts["protect"] + counts["set-null"],
        },
        "roots": roots,
    }
    return preview, ordered_items, set_null_items


async def build_delete_preview(
        session: AsyncSession,
        registry: Registry,
        model_config: ModelConfig,
        items: list[Any],
) -> dict[str, Any]:
    preview, _ordered_items, _set_null_items = await build_delete_plan(session, registry, model_config, items)
    return preview


async def _build_dependency_graph(
        *,
        session: AsyncSession,
        registry: Registry,
        root_model_config: ModelConfig,
        root_items: list[Any],
        set_null_items: list[tuple[Any, list[str]]],
        set_null_index: dict[NodeKey, int],
) -> dict[NodeKey, list[_PreviewEdge]]:
    edges_by_parent: dict[NodeKey, list[_PreviewEdge]] = defaultdict(list)
    expanded_keys: set[NodeKey] = set()
    pending_groups: dict[str, tuple[ModelConfig, list[Any]]] = {}
    _append_pending_items(pending_groups, root_model_config, root_items)

    while pending_groups:
        current_groups = pending_groups
        pending_groups = {}

        for current_model_config, current_items in current_groups.values():
            items_to_expand = [
                item
                for item in current_items
                if (current_model_config.model, get_pk_value(current_model_config, item)) not in expanded_keys
            ]
            if not items_to_expand:
                continue

            parent_ids = [get_pk_value(current_model_config, item) for item in items_to_expand]
            parent_id_set = set(parent_ids)
            for item in items_to_expand:
                expanded_keys.add((current_model_config.model, get_pk_value(current_model_config, item)))

            for dependent_relation in _iter_dependent_relations(current_model_config.model):
                related_items = await _fetch_dependent_items(session, dependent_relation, parent_id_set)
                related_config = registry.find_by_model(dependent_relation.model)

                for related_item in related_items:
                    matching_parent_ids = _resolve_matching_parent_ids(
                        related_item,
                        dependent_relation.attribute_names,
                        parent_id_set,
                    )
                    if not matching_parent_ids:
                        continue

                    preview_edge = _build_preview_edge(
                        model_config=related_config,
                        related_item=related_item,
                        dependent_relation=dependent_relation,
                    )
                    for parent_id in matching_parent_ids:
                        parent_key = (current_model_config.model, parent_id)
                        edges_by_parent[parent_key].append(preview_edge)

                    if related_config is None:
                        continue
                    if dependent_relation.effect == "set-null":
                        _append_set_null_item(
                            set_null_items,
                            set_null_index,
                            related_item,
                            dependent_relation.attribute_names,
                        )
                        continue
                    _append_pending_items(pending_groups, related_config, [related_item])

    return dict(edges_by_parent)


def _build_preview_edge(
        *,
        model_config: ModelConfig | None,
        related_item: Any,
        dependent_relation: _DependentRelation,
) -> _PreviewEdge:
    if model_config is None:
        return _PreviewEdge(
            relation_name=dependent_relation.relation_name,
            effect=dependent_relation.effect,
            model_config=None,
            item=related_item,
            model_title=dependent_relation.model.__name__,
            label=_get_fallback_label(related_item),
        )

    return _PreviewEdge(
        relation_name=dependent_relation.relation_name,
        effect=dependent_relation.effect,
        model_config=model_config,
        item=related_item,
        model_title=_get_model_title(model_config),
        label=get_display_value(model_config, related_item),
    )


def _append_pending_items(
        pending_groups: dict[str, tuple[ModelConfig, list[Any]]],
        model_config: ModelConfig,
        items: list[Any],
) -> None:
    if not items:
        return
    slug = model_config.slug
    if slug is None:
        raise ValueError(f"Model '{model_config.model.__name__}' must have a resolved slug.")
    existing = pending_groups.get(slug)
    if existing is None:
        pending_groups[slug] = (model_config, list(items))
        return
    existing[1].extend(items)


def _append_set_null_item(
        set_null_items: list[tuple[Any, list[str]]],
        set_null_index: dict[NodeKey, int],
        item: Any,
        attribute_names: list[str],
) -> None:
    item_key = (type(item), _get_raw_pk_value(item))
    existing_index = set_null_index.get(item_key)
    if existing_index is None:
        set_null_index[item_key] = len(set_null_items)
        set_null_items.append((item, list(attribute_names)))
        return

    existing_attributes = set_null_items[existing_index][1]
    for attribute_name in attribute_names:
        if attribute_name not in existing_attributes:
            existing_attributes.append(attribute_name)


async def _fetch_dependent_items(
        session: AsyncSession,
        dependent_relation: _DependentRelation,
        parent_ids: set[Any],
) -> list[Any]:
    if not parent_ids:
        return []
    return list((await session.execute(_build_dependent_query(parent_ids, dependent_relation))).scalars().unique())


def _build_dependent_query(parent_ids: set[Any], dependent_relation: _DependentRelation) -> Any:
    return select(dependent_relation.model).where(
        or_(
            *[
                getattr(dependent_relation.model, attribute_name).in_(parent_ids)
                for attribute_name in dependent_relation.attribute_names
            ]
        ),
    )


def _resolve_matching_parent_ids(item: Any, attribute_names: list[str], parent_ids: set[Any]) -> list[Any]:
    matching_parent_ids: list[Any] = []
    for attribute_name in attribute_names:
        parent_id = getattr(item, attribute_name, None)
        if parent_id in parent_ids and parent_id not in matching_parent_ids:
            matching_parent_ids.append(parent_id)
    return matching_parent_ids


def _build_preview_node(
        *,
        registry: Registry,
        model_config: ModelConfig,
        item: Any,
        relation_name: str | None,
        effect: DeleteEffect,
        path_seen: set[NodeKey],
        edges_by_parent: dict[NodeKey, list[_PreviewEdge]],
        ordered_items: list[Any],
        ordered_keys: set[NodeKey],
) -> dict[str, Any]:
    item_id = get_pk_value(model_config, item)
    node_key = (type(item), item_id)
    if node_key in path_seen:
        return _build_node_payload(
            model_slug=model_config.slug,
            model_title=_get_model_title(model_config),
            relation_name=relation_name,
            item_id=item_id,
            label=get_display_value(model_config, item),
            children=[],
            effect=effect,
        )

    next_path_seen = set(path_seen)
    next_path_seen.add(node_key)
    children: list[dict[str, Any]] = []
    for edge in edges_by_parent.get(node_key, []):
        if edge.model_config is None:
            children.append(
                _build_node_payload(
                    model_slug=None,
                    model_title=edge.model_title,
                    relation_name=edge.relation_name,
                    item_id=_get_raw_pk_value(edge.item),
                    label=edge.label,
                    children=[],
                    effect=edge.effect,
                ),
            )
            continue

        if edge.effect == "set-null":
            children.append(
                _build_node_payload(
                    model_slug=edge.model_config.slug,
                    model_title=edge.model_title,
                    relation_name=edge.relation_name,
                    item_id=get_pk_value(edge.model_config, edge.item),
                    label=edge.label,
                    children=[],
                    effect="set-null",
                ),
            )
            continue

        children.append(
            _build_preview_node(
                registry=registry,
                model_config=edge.model_config,
                item=edge.item,
                relation_name=edge.relation_name,
                effect=edge.effect,
                path_seen=next_path_seen,
                edges_by_parent=edges_by_parent,
                ordered_items=ordered_items,
                ordered_keys=ordered_keys,
            ),
        )

    if effect == "delete" and node_key not in ordered_keys:
        ordered_items.append(item)
        ordered_keys.add(node_key)

    return _build_node_payload(
        model_slug=model_config.slug,
        model_title=_get_model_title(model_config),
        relation_name=relation_name,
        item_id=item_id,
        label=get_display_value(model_config, item),
        children=children,
        effect=effect,
    )


def _iter_dependent_relations(model: type[Any]) -> list[_DependentRelation]:
    cached_relations = _DEPENDENT_RELATIONS_CACHE.get(model)
    if cached_relations is not None:
        return cached_relations

    mapper = sa_inspect(model)
    pk_column = mapper.primary_key[0]
    relations: list[_DependentRelation] = []

    for candidate_mapper in mapper.registry.mappers:
        candidate_model = candidate_mapper.class_
        if candidate_model is model:
            continue

        columns = [
            column
            for column in candidate_mapper.columns
            if any(
                foreign_key.column.table is pk_column.table and foreign_key.column.key == pk_column.key
                for foreign_key in column.foreign_keys
            )
        ]
        if not columns:
            continue

        relations.append(
            _DependentRelation(
                model=candidate_model,
                relation_name=", ".join(column.key for column in columns),
                columns=columns,
                attribute_names=[_get_attribute_name_for_column(candidate_mapper, column) for column in columns],
                effect=_resolve_dependent_effect(model, candidate_model, columns),
            ),
        )

    _DEPENDENT_RELATIONS_CACHE[model] = relations
    return relations


def _resolve_dependent_effect(
        parent_model: type[Any],
        candidate_model: type[Any],
        columns: list[Any],
) -> DeleteEffect:
    relationship_effect = _get_relationship_effect(parent_model, candidate_model, columns)
    if relationship_effect is not None:
        return relationship_effect
    if any(_column_has_ondelete(column, "CASCADE") for column in columns):
        return "delete"
    if any(_column_has_ondelete(column, "SET NULL") for column in columns):
        return "set-null"
    return "protect"


def _get_relationship_effect(
        parent_model: type[Any],
        candidate_model: type[Any],
        columns: list[Any],
) -> DeleteEffect | None:
    parent_mapper = sa_inspect(parent_model)
    column_keys = {column.key for column in columns}

    for relationship in parent_mapper.relationships:
        if relationship.mapper.class_ is not candidate_model:
            continue
        if relationship.secondary is not None or relationship.viewonly:
            continue
        if relationship.direction is not RelationshipDirection.ONETOMANY:
            continue

        relationship_column_keys = {remote.key for _local, remote in relationship.local_remote_pairs}
        if relationship_column_keys and not relationship_column_keys.issubset(column_keys):
            continue
        if _relationship_cascades_delete(relationship.cascade):
            return "delete"

    return None


def _relationship_cascades_delete(cascade: Any) -> bool:
    return bool(getattr(cascade, "delete", False) or getattr(cascade, "delete_orphan", False))


def _column_has_ondelete(column: Any, value: str) -> bool:
    return any(
        isinstance(getattr(foreign_key, "ondelete", None), str) and foreign_key.ondelete.upper() == value
        for foreign_key in column.foreign_keys
    )


def _get_attribute_name_for_column(candidate_mapper: Any, column: Any) -> str:
    return candidate_mapper.get_property_by_column(column).key


def _count_effects(roots: list[dict[str, Any]]) -> dict[DeleteEffect, int]:
    counts: dict[DeleteEffect, int] = {"delete": 0, "protect": 0, "set-null": 0}
    for root in roots:
        for child in root.get("children", []):
            _accumulate_effects(child, counts)
    return counts


def _accumulate_effects(node: dict[str, Any], counts: dict[DeleteEffect, int]) -> None:
    effect = node.get("effect", "delete")
    counts[effect] += 1
    for child in node.get("children", []):
        _accumulate_effects(child, counts)


def _build_node_payload(
        *,
        model_slug: str | None,
        model_title: str,
        relation_name: str | None,
        item_id: Any,
        label: str,
        children: list[dict[str, Any]],
        effect: DeleteEffect,
) -> dict[str, Any]:
    return {
        "model_slug": model_slug,
        "model_title": model_title,
        "relation_name": relation_name,
        "id": serialize_scalar(item_id),
        "label": label,
        "effect": effect,
        "children": children,
    }


def _get_raw_pk_value(item: Any) -> Any:
    mapper = sa_inspect(type(item))
    pk_column = mapper.primary_key[0]
    return getattr(item, pk_column.key)


def _get_fallback_label(item: Any) -> str:
    item_id = serialize_scalar(_get_raw_pk_value(item))
    for candidate in ("name", "title", "username", "slug", "email", "code_name"):
        if hasattr(item, candidate):
            raw_value = getattr(item, candidate)
            if raw_value is not None:
                return str(raw_value)
    return f"{type(item).__name__} #{item_id}"


def _get_model_title(model_config: ModelConfig) -> str:
    return model_config.title or model_config.slug or model_config.model.__name__


@dataclass(slots=True)
class _DependentRelation:
    model: type[Any]
    relation_name: str
    columns: list[Any]
    attribute_names: list[str]
    effect: DeleteEffect


@dataclass(slots=True)
class _PreviewEdge:
    relation_name: str | None
    effect: DeleteEffect
    model_config: ModelConfig | None
    item: Any
    model_title: str
    label: str



from xladmin.introspection_fields import (
    get_all_field_names,
    get_column_names,
    get_create_fields,
    get_display_value,
    get_mapper_field_names,
    get_pk_field_name,
    get_pk_value,
    get_relationship_names,
    get_sort_column_name,
    get_update_fields,
    get_visible_detail_fields,
    get_visible_list_fields,
    pk_is_generated,
)
from xladmin.introspection_meta import (
    get_list_filter_input_kind,
    get_list_filter_options,
    get_model_blocks_meta,
    get_model_meta,
    has_relation_choice_endpoint,
    list_filter_has_choices,
)
from xladmin.introspection_values import convert_boolean_scalar, convert_value_for_column, get_field_value

__all__ = (
    "convert_boolean_scalar",
    "convert_value_for_column",
    "get_all_field_names",
    "get_column_names",
    "get_create_fields",
    "get_display_value",
    "get_field_value",
    "get_list_filter_input_kind",
    "get_list_filter_options",
    "get_mapper_field_names",
    "get_model_blocks_meta",
    "get_model_meta",
    "get_pk_field_name",
    "get_pk_value",
    "pk_is_generated",
    "get_relationship_names",
    "get_sort_column_name",
    "get_update_fields",
    "get_visible_detail_fields",
    "get_visible_list_fields",
    "has_relation_choice_endpoint",
    "list_filter_has_choices",
)



from __future__ import annotations

import inspect
from typing import Any, Literal
from uuid import UUID

from fastapi import HTTPException, status
from sqlalchemy import func, or_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.inspection import inspect as sa_inspect
from sqlalchemy.orm import selectinload

from xladmin.config import ModelConfig
from xladmin.i18n import translate
from xladmin.introspection import (
    convert_boolean_scalar,
    convert_value_for_column,
    get_pk_field_name,
    get_sort_column_name,
)


def apply_ordering(model_config: ModelConfig, query: Any, sort: str | None = None) -> Any:
    ordering_items = (
        [item.strip() for item in (sort or "").split(",") if item.strip()]
        if sort
        else list(model_config.ordering)
    )
    if not ordering_items:
        return query

    for item in ordering_items:
        descending = item.startswith("-")
        field_name = item[1:] if descending else item
        sort_column_name = get_sort_column_name(model_config, field_name)
        if sort_column_name is None:
            continue
        column = getattr(model_config.model, sort_column_name)
        query = query.order_by(column.desc() if descending else column.asc())
    return query


def apply_search(model_config: ModelConfig, query: Any, q: str | None, session: AsyncSession) -> Any:
    if not q:
        return query
    if model_config.search_query_builder is not None:
        return model_config.search_query_builder(query, q, session)
    search_fields = model_config.search_fields or tuple(
        column_name
        for column_name, column in sa_inspect(model_config.model).columns.items()
        if getattr(column.type, "python_type", None) is str
    )
    conditions = [getattr(model_config.model, field_name).ilike(f"%{q}%") for field_name in search_fields]
    if not conditions:
        return query
    return query.where(or_(*conditions))


async def apply_list_filters(
        model_config: ModelConfig,
        query: Any,
        values: dict[str, str],
        session: AsyncSession,
        user: Any,
) -> Any:
    if not values:
        return query

    mapper = sa_inspect(model_config.model)
    for list_filter in model_config.list_filters:
        raw_value = values.get(list_filter.slug)
        if raw_value in (None, ""):
            continue

        filter_value = str(raw_value)
        filter_values = parse_list_filter_values(filter_value, list_filter)
        parsed_filter_values = [
            list_filter.value_parser(item) if list_filter.value_parser is not None else item
            for item in filter_values
        ]
        parsed_value = parsed_filter_values if list_filter_is_multiple(list_filter) else parsed_filter_values[0]
        option_config = next((item for item in list_filter.options if item.value == filter_value), None)
        filter_handler = option_config.filter_handler if option_config is not None else list_filter.filter_handler
        if filter_handler is not None:
            custom_query = filter_handler(query, parsed_value, session, user)
            query = await custom_query if inspect.isawaitable(custom_query) else custom_query
            continue

        if list_filter.field_name is None:
            continue

        if list_filter.field_name in mapper.relationships:
            relationship = mapper.relationships[list_filter.field_name]
            related_pk_name = relationship.mapper.primary_key[0].key
            related_pk_attr = getattr(relationship.mapper.class_, related_pk_name)
            if relationship.uselist:
                normalized_values = [convert_pk(item) for item in parsed_filter_values]
                if not normalized_values:
                    continue
                query = query.where(
                    getattr(model_config.model, list_filter.field_name).any(
                        related_pk_attr.in_(normalized_values),
                    ),
                )
                continue
            local_columns = list(relationship.local_columns)
            if not local_columns:
                continue
            normalized_values = [convert_pk(item) for item in parsed_filter_values]
            if not normalized_values:
                continue
            local_attr = getattr(model_config.model, local_columns[0].key)
            query = query.where(local_attr.in_(normalized_values))
            continue

        if list_filter.field_name not in mapper.columns:
            continue

        column = mapper.columns[list_filter.field_name]
        column_attr = getattr(model_config.model, list_filter.field_name)
        input_kind = list_filter.input_kind or resolve_list_filter_input_kind(column, list_filter)
        if input_kind == "boolean":
            normalized_values = [convert_boolean_scalar(item) for item in parsed_filter_values]
            query = query.where(column_attr.in_(normalized_values))
            continue
        if input_kind == "text":
            query = query.where(column_attr.ilike(f"%{parsed_value}%"))
            continue
        normalized_values = [convert_value_for_column(column, item) for item in parsed_filter_values]
        query = query.where(column_attr.in_(normalized_values))

    return query


def apply_eager_loads(model_config: ModelConfig, query: Any, *, mode: Literal["list", "detail"]) -> Any:
    mapper = sa_inspect(model_config.model)
    relation_names = set(mapper.relationships.keys())
    configured_fields = (
        tuple(model_config.list_display or ())
        if mode == "list"
        else tuple(dict.fromkeys((*(model_config.detail_fields or ()), *model_config.fields.keys())))
    )
    relation_fields = [field_name for field_name in configured_fields if field_name in relation_names]
    for field_name in dict.fromkeys(relation_fields):
        query = query.options(selectinload(getattr(model_config.model, field_name)))
    return query


def resolve_list_filter_input_kind(column: Any, list_filter: Any) -> str:
    if list_filter.options:
        return "select"
    try:
        python_type = column.type.python_type
    except NotImplementedError:
        return "text"
    if python_type is bool:
        return "boolean"
    return "text"


def resolve_relation_model(model_config: ModelConfig, field_name: str, *, locale: str | None = None) -> type[Any]:
    configured_model = model_config.get_field_config(field_name).relation_model
    if configured_model is not None:
        return configured_model

    mapper = sa_inspect(model_config.model)
    if field_name in mapper.relationships:
        return mapper.relationships[field_name].mapper.class_
    column = mapper.columns[field_name]
    foreign_key = next(iter(column.foreign_keys), None)
    if foreign_key is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=translate(locale, "field_has_no_relation_choices"),
        )
    remote_table = foreign_key.column.table
    for registry_mapper in mapper.registry.mappers:
        model = registry_mapper.class_
        if getattr(model, "__table__", None) is remote_table:
            return model
    raise HTTPException(
        status_code=status.HTTP_400_BAD_REQUEST,
        detail=translate(locale, "related_model_not_found"),
    )


def resolve_list_filter_relation_model(
        model_config: ModelConfig,
        list_filter: Any,
        *,
        locale: str | None = None,
) -> type[Any] | None:
    if list_filter.relation_model is not None:
        return list_filter.relation_model
    if list_filter.field_name is None:
        return None

    mapper = sa_inspect(model_config.model)
    if list_filter.field_name in mapper.relationships:
        return mapper.relationships[list_filter.field_name].mapper.class_
    if list_filter.field_name in mapper.columns and mapper.columns[list_filter.field_name].foreign_keys:
        return resolve_relation_model(model_config, list_filter.field_name, locale=locale)
    return None


def resolve_label_field(model: type[Any]) -> str:
    for candidate in ("name", "title", "username", "slug", "email", "code_name"):
        if hasattr(model, candidate):
            return candidate
    return sa_inspect(model).primary_key[0].key


def convert_pk(value: Any) -> Any:
    if isinstance(value, UUID):
        return value
    if isinstance(value, int):
        return value
    if isinstance(value, str):
        if value.isdigit():
            return int(value)
        try:
            return UUID(value)
        except ValueError:
            return value
    return value


def parse_list_filter_values(value: str, list_filter: Any) -> list[str]:
    if list_filter_is_multiple(list_filter):
        return [item.strip() for item in value.split(",") if item.strip()]
    return [value]


def list_filter_is_multiple(list_filter: Any) -> bool:
    return bool(
        getattr(list_filter, "multiple", False)
        or getattr(list_filter, "input_kind", None) == "select-multiple"
    )


async def apply_scoped_query(query: Any, model_config: ModelConfig, session: AsyncSession, user: Any) -> Any:
    if model_config.query_for_list is None:
        return query
    custom_query = model_config.query_for_list(query, session, user)
    return await custom_query if inspect.isawaitable(custom_query) else custom_query


async def build_model_query(
        session: AsyncSession,
        model_config: ModelConfig,
        user: Any,
        *,
        mode: Literal["list", "detail"],
) -> Any:
    query = select(model_config.model)
    query = await apply_scoped_query(query, model_config, session, user)
    return apply_eager_loads(model_config, query, mode=mode)


async def get_item_by_pk(
        session: AsyncSession,
        model_config: ModelConfig,
        item_id: str,
        user: Any,
        *,
        mode: Literal["list", "detail"] = "detail",
) -> Any:
    pk_attr = getattr(model_config.model, get_pk_field_name(model_config))
    query = await build_model_query(session, model_config, user, mode=mode)
    query = query.where(pk_attr == convert_pk(item_id))
    return (await session.execute(query)).scalars().unique().one_or_none()


async def get_items_by_ids(
        session: AsyncSession,
        model_config: ModelConfig,
        ids: list[Any],
        user: Any,
        *,
        mode: Literal["list", "detail"] = "detail",
) -> list[Any]:
    normalized_ids = [convert_pk(item_id) for item_id in ids]
    if not normalized_ids:
        return []
    pk_attr = getattr(model_config.model, get_pk_field_name(model_config))
    query = await build_model_query(session, model_config, user, mode=mode)
    query = query.where(pk_attr.in_(normalized_ids))
    items = list((await session.execute(query)).scalars().unique())
    items_by_pk = {getattr(item, pk_attr.key): item for item in items}
    return [items_by_pk[item_id] for item_id in normalized_ids if item_id in items_by_pk]


async def count_total_items(session: AsyncSession, model_config: ModelConfig, query: Any) -> int:
    pk_name = get_pk_field_name(model_config)
    query_subquery = query.order_by(None).subquery()
    total_query = select(func.count(func.distinct(getattr(query_subquery.c, pk_name))))
    return int((await session.execute(total_query)).scalar_one())
